/*
 * $Id: //depot/prod/spl/main/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/command/MountSmBackupCommand.java#4 $
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.snapcenter.cli.command;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.Parameters;
import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.smcore.plugin.contracts.ArrayOfSmBackupComponentInfo;
import com.netapp.smcore.plugin.contracts.ArrayOfSmOracleASMDiskGroup;
import com.netapp.smcore.plugin.contracts.SmBackup;
import com.netapp.smcore.plugin.contracts.SmBackupComponentInfo;
import com.netapp.smcore.plugin.contracts.SmCreateCloneRequest;
import com.netapp.smcore.plugin.contracts.SmCreateCloneResponse;
import com.netapp.smcore.plugin.contracts.SmGetBackupDetailsResponse;
import com.netapp.smcore.plugin.contracts.SmOracleASMDiskGroup;
import com.netapp.smcore.plugin.contracts.SmOracleBackupComponentInfo;
import com.netapp.smcore.plugin.contracts.SmOracleBackupMetadata;
import com.netapp.smcore.plugin.contracts.SmRequest;
import com.netapp.smcore.plugin.contracts.SmResponse;
import com.netapp.snapcenter.cli.command.parser.CustomStringToMapConverter;
import com.netapp.snapcenter.cli.command.parser.NoParameterSplitter;
import com.netapp.snapcenter.cli.command.validation.SmsBackupValidationHelper;
import com.netapp.snapcenter.cli.command.validation.SmsCommandArgumentsValidationUtil;
import com.netapp.snapcenter.cli.command.validation.SmsCommandRequestValidationHelper;
import com.netapp.snapcenter.cli.contracts.PluginCode;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.helper.SmOracleDatabaseConfigurationHelper;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;
import com.netapp.snapcenter.cli.utils.ProgressBar;
import com.netapp.snapcenter.cli.utils.ProgressBarUtil;
import com.netapp.snapcenter.cli.utils.SmoCloneCliUtil;
import com.netapp.spl.common.restclient.config.SMSRestConfiguration;
import com.netapp.spl.common.restclient.context.SMSRestInvocationContext;
import com.netapp.spl.common.utils.HostInfoUtil;

@Parameters(resourceBundle = "snapcenter-cli", commandDescriptionKey = "cli.mountsmbackup.command.description")
public class MountSmBackupCommand extends AbstractSMSCommand {

    private ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(MountSmBackupCommand.class);

    private static final String SMO_MOUNT_BACKUP_LOCATION_FORMAT = "/var/opt/snapcenter/sco/backup_mount/%s/%s";

	private static final String SMO_ASM_MOUNT_BACKUP_LOCATION_FORMAT = "+%s_%s_%s";

	private static final String SMO_ASM_MOUNT_BACKUP_PREFIX = "SC_";

	private static final int SMO_ASM_DISK_GROUP_LENGTH = 15;

    private static final String COMMANDNAME = MessageBundle
            .localizeMessage(CLIMsgKey.CLI_MOUNTSMBACKUP_COMMAND_NAME);

    @Parameter(names = { "-HostName" }, descriptionKey = "cli.mountsmbackup.command.hostName.option.description")
    private String hostName;

    @Parameter(names = { "-BackupName" }, descriptionKey = "cli.mountsmbackup.command.backupName.option.description", required = true)
    private String backupName;

    @Parameter(names = { "-AppObjectId" }, descriptionKey = "cli.mountsmbackup.command.appObjectId.option.description", required = true)
    private String appObjectId;

    @Parameter(names = { "-AppPluginCode" }, descriptionKey = "cli.mountsmbackup.command.appPluginCode.option.description")
    private PluginCode appPluginCode = PluginCode.SCO;

    @Parameter(names = { "-ASMCredentialName" }, descriptionKey = "cli.mountsmbackup.command.asmRunasName.option.description")
    private String asmRunAsName;

    @Parameter(names = { "-AsmPort" }, descriptionKey = "cli.mountsmbackup.command.asmPort.option.description")
    private int asmPort = SmOracleDatabaseConfigurationHelper.DATABASE_OR_ASM_DEFAULT_PORT;

    @Parameter(names = { "-SecondaryLocator" }, descriptionKey = "cli.mountsmbackup.command.secondaryLocator.option.description", converter = CustomStringToMapConverter.class, splitter = NoParameterSplitter.class)
    private List<Map<String, String>> secondaryLocators;

    @Parameter(names = { "-WaitForCompletion" }, descriptionKey = "cli.mountsmbackup.command.waitForCompletion.option.description")
    private boolean waitForCompletion;

    private boolean fileSystemMount = true;

    private String mountPath;

	private boolean isDataBackup;

	private boolean mountFromSecondary;

	private boolean ContainsNfsOrSanMountPoint;

	private boolean isAsmDatabase;

	private ArrayList<String> mountPathList = new ArrayList<String>();

    @Override
    public String getCommandName() {
        return COMMANDNAME;
    }

    @Override
	protected void processCommandArguments() throws CLIException {

		ProgressBar progressBar = ProgressBarUtil
				.createAndStartProgressBar(MessageBundle
						.localizeMessage(CLIMsgKey.CLI_PROGRESSBAR_FETCH_MOUNTPATH));

		try {

			SmsBackupValidationHelper smsBackupValidationHelper = new SmsBackupValidationHelper(
					getSnapCenterServerServiceHelper());

			setMountFromSecondary(false);

			if (!SmsCommandArgumentsValidationUtil
					.isValidAppObjectId(appObjectId)) {
				throw new CLIException(
						MessageBundle
								.localizeMessage(CLIMsgKey.CLI_COMMAND_PARAMETER_APP_OBJECT_ID_NOT_VALID));
			}

			if (CLIUtils.isNotNullAndNotEmptyCollection(secondaryLocators)) {
				SmsCommandArgumentsValidationUtil
						.validateSecondaryLocators(secondaryLocators);
				setMountFromSecondary(true);
				CLIUtils.writeToLogAndConsole(
						MessageBundle
								.localizeMessage(CLIMsgKey.CLI_MOUNTSMBACKUP_WILL_BE_PERFORMED_FROM_SECONDARY),
						logger, LogLevel.INFO);
			}

			switch (appPluginCode) {
			case SCO:

				if (!isValidOracleBackup(smsBackupValidationHelper,
						mountFromSecondary)) {
					if (mountFromSecondary) {
						throw new CLIException(
								MessageBundle
										.localizeMessage(
												CLIMsgKey.CLI_MOUNTSMBACKUP_NOT_VALID_ORACLE_SECONDARY_BACKUP_NAME_OR_RESOURCE_NOT_IN_BACKUP,
												getBackupName(), appObjectId));
					} else {
						throw new CLIException(
								MessageBundle
										.localizeMessage(
												CLIMsgKey.CLI_MOUNTSMBACKUP_NOT_VALID_ORACLE_BACKUP_NAME_OR_RESOURCE_NOT_IN_BACKUP,
												getBackupName(), appObjectId));
					}
				}

				String dbName = StringUtils.split(appObjectId, "\\")[1];

				String backupId = getValidOracleBackupId(
						smsBackupValidationHelper, mountFromSecondary);

				if (StringUtils.isBlank(backupId)) {
					throw new CLIException(
							MessageBundle
									.localizeMessage(
											CLIMsgKey.CLI_MOUNTSMBACKUP_NOT_VALID_ORACLE_BACKUP_NAME_OR_RESOURCE_NOT_IN_BACKUP,
											getBackupName(), appObjectId));
				}

				setDataBackup(smsBackupValidationHelper
						.checkIfValidOracleDataBackup(backupId, backupName,
								appObjectId, mountFromSecondary));

				buildMountPath(dbName, backupId, appObjectId);

				if (getAsmRunAsName() != null
						&& StringUtils.isBlank(getAsmRunAsName())) {
					// validate ASM auth
					throw new CLIException(
							MessageBundle
									.localizeMessage(CLIMsgKey.CLI_MOUNTSMBACKUP_INVALID_ASM_AUTH_RUNAS));
				}

				if (!CLIUtils.isValidPortNumber(getAsmPort())) {
					// set ASM port number
					throw new CLIException(MessageBundle.localizeMessage(
							CLIMsgKey.CLI_MOUNTSMBACKUP_INVALID_ASM_AUTH_PORT,
							getAsmPort()));
				}

				ProgressBarUtil.stopProgressBar(progressBar);

				displayMountPath();
	
				break;

			default:
				throw new CLIException(MessageBundle.localizeMessage(
						CLIMsgKey.CLI_COMMAND_NOT_SUPPORTED_FOR_PLUGIN,
						appPluginCode.name()));
			}

		} catch (CLIException e) {
			throw e;
		} finally {
			ProgressBarUtil.stopProgressBar(progressBar);
		}

		if (StringUtils.isBlank(getHostName())) {
			setHostName(HostInfoUtil.getCanonicalHostName());
			String usingDefaultHost = MessageBundle.localizeMessage(
					CLIMsgKey.CLI_MOUNTSMBACKUP_USING_LOCALHOST_AS_DEFAULT,
					getHostName());
			CLIUtils.writeToLogAndConsole(usingDefaultHost, logger,
					LogLevel.INFO);
		}
	}

	private boolean isValidOracleBackup(
			SmsBackupValidationHelper smsBackupValidationHelper,
			boolean mountFromSecondary) throws CLIException {
		boolean validOracleBackup = false;
		try {
			// Validate the specified backup
			validOracleBackup = smsBackupValidationHelper
					.checkIfValidOracleBackup(null, backupName, appObjectId,
							mountFromSecondary);
		} catch (CLIException e) {
			throw e;
		} 
		return validOracleBackup;
	}

	private String getValidOracleBackupId(
			SmsBackupValidationHelper smsBackupValidationHelper,
			boolean mountFromSecondary) throws CLIException {

		String validOracleBackupId = null;
		try {
			// Validate the specified backup
			validOracleBackupId = smsBackupValidationHelper
					.fetchBackupIdIfItIsValidOracleBackup(backupName,
							appObjectId, mountFromSecondary);
		} catch (CLIException e) {
			throw e;
		}
		logger.debug("BackupId of the " + appObjectId + " is " + validOracleBackupId );
		return validOracleBackupId;
	}

	private void buildMountPath(String dbName, String backupId,
			String appObjectId) throws CLIException {
		try {
			SmsCommandRequestValidationHelper smsCommandRequestValidationHelper = new SmsCommandRequestValidationHelper(
					super.getSnapCenterServerServiceHelper());
			SmGetBackupDetailsResponse smGetBackupDetailsResponse = smsCommandRequestValidationHelper
					.getBackupDetailsFromServer(appObjectId, backupId);

			if (StringUtils.isNotBlank(smGetBackupDetailsResponse.getResult()
					.getMessage())) {
				throw new CLIException(smGetBackupDetailsResponse.getResult()
						.getMessage());
			}
			logger.debug("Size of SmBackup "
					+ smGetBackupDetailsResponse.getBackups().getSmBackup()
							.size());
			String diskGroupName = null;

			for (SmBackup listOfSmBackup : smGetBackupDetailsResponse
					.getBackups().getSmBackup()) {
				ArrayOfSmBackupComponentInfo arrayOfSmBackupComponentInfo = listOfSmBackup
						.getBackupComponents();
				SmOracleBackupComponentInfo smOracleBackupComponentInfo = (SmOracleBackupComponentInfo) arrayOfSmBackupComponentInfo
						.getSmBackupComponentInfo().iterator().next();

				if ((smOracleBackupComponentInfo.getSmOracleDatabaseInstance() != null)
						&& (smOracleBackupComponentInfo
								.getSmOracleDatabaseInstance().getASMDetails() != null)
						&& StringUtils.isNotBlank(smOracleBackupComponentInfo
								.getSmOracleDatabaseInstance().getASMDetails()
								.getASMInstanceName())) {
					int diskGroupSize = smOracleBackupComponentInfo
							.getSmOracleDatabaseInstance().getASMDetails()
							.getASMDiskGroups().getSmOracleASMDiskGroup()
							.size();
					if (diskGroupSize > 0) {
						logger.debug("Size of ASM disk group " + diskGroupSize);

						SmOracleBackupMetadata smOracleBackupMetadata = smsCommandRequestValidationHelper
								.getSmOracleBackupMetadata(appObjectId,
										backupId, backupName);
						isAsmDatabase = smOracleBackupMetadata
								.isIsAsmDatabase();
						ContainsNfsOrSanMountPoint = smOracleBackupMetadata
								.isContainsNfsOrSanMountPoint();
	
						logger.debug("building mount path, isAsmDatabase: " + isAsmDatabase
								+ " ContainsNfsOrSanMountPoint:"
								+ ContainsNfsOrSanMountPoint);

						for (SmOracleASMDiskGroup smOracleASMDiskGroup : smOracleBackupComponentInfo
								.getSmOracleDatabaseInstance().getASMDetails()
								.getASMDiskGroups().getSmOracleASMDiskGroup()) {

							if (StringUtils.isBlank(smOracleASMDiskGroup
									.getName())) {
								throw new CLIException(
										MessageBundle
												.localizeMessage(
														CLIMsgKey.CLI_MOUNTSMBACKUP_ASM_DISKGROUP_EMPTY_ERROR,
														dbName));
							}

							if (smOracleASMDiskGroup.getName().length() > SMO_ASM_DISK_GROUP_LENGTH) {

								diskGroupName = SMO_ASM_MOUNT_BACKUP_PREFIX
										+ SmoCloneCliUtil
												.getHashCode(smOracleASMDiskGroup
														.getName());
							} else {

								diskGroupName = smOracleASMDiskGroup.getName();
							}

							logger.debug("Setting mount path for asm db "
									+ dbName + " as " + diskGroupName);

							mountPathList.add(String.format(
									SMO_ASM_MOUNT_BACKUP_LOCATION_FORMAT,
									diskGroupName, smOracleBackupComponentInfo
											.getSmOracleDatabaseInstance()
											.getASMDetails()
											.getASMInstanceName(), backupId));

						}
					}

				}
			}

			/*
			 * If ASM DB is only on SAN then set the mount path to disk groups
			 * names and for rest of the cases set to path
			 * SMO_MOUNT_BACKUP_LOCATION_FORMAT.
			 */
			if (!ContainsNfsOrSanMountPoint && isAsmDatabase) {
				setMountPath(StringUtils.join(mountPathList, ","));
				logger.debug("Setting mount path for asm db: " + dbName
						+ "over raw device: " + " is: " + getMountPath());
			} else {
				setMountPath(String.format(SMO_MOUNT_BACKUP_LOCATION_FORMAT,
						getBackupName(), dbName));
				logger.debug("Setting mount path db: " + dbName + " is: "
						+ getMountPath());
			}

		} catch (CLIException e) {
			throw e;
		}

	}

	private void displayMountPath() {

		String mountPath = getMountPath();
		/*
		 * If ASM DB is on SAN and NFS then show disk groups names and NFS mount
		 * path also, since ASM over NFS will be mounted under the NFS mount
		 * path.
		 * In other cases just printing the mount path is set. 
		 */
		if (ContainsNfsOrSanMountPoint && isAsmDatabase) {
			mountPath = StringUtils.join(mountPathList, ",") + ","
					+ getMountPath();
		}
		logger.debug("Displayed mount path: " + mountPath);
		CLIUtils.writeToLogAndConsole(MessageBundle.localizeMessage(
				CLIMsgKey.CLI_MOUNTSMBACKUP_DISPLAY_MOUNT_PATH, mountPath),
				logger, LogLevel.INFO);
	}

    @Override
    protected <T> void pollForJobStatus(SmResponse smResponse)
            throws CLIException {

        doPollForJobStatus(smResponse, isWaitForCompletion());
    }

    @Override
    protected <T extends SmRequest> SMSRestInvocationContext<?, ?> getCommandSpecificSMSRestInvocationContext(
            T cmdRequest) {
        SMSRestInvocationContext<SmCreateCloneRequest, SmCreateCloneResponse> smsRestInvocationContext = new SMSRestInvocationContext<>();
        smsRestInvocationContext.setRequest((SmCreateCloneRequest) cmdRequest);
        smsRestInvocationContext.setResponseClass(SmCreateCloneResponse.class);
        smsRestInvocationContext
                .setServiceUrlPath(SMSRestConfiguration.SMS_CREATE_CLONE_SERVICE_PATH);
        return smsRestInvocationContext;
    }

    public void setHostName(String hostName) {
        this.hostName = hostName;
    }

    public String getHostName() {
        return hostName;
    }

    public String getBackupName() {
        return backupName;
    }

    public void setBackupName(String backupName) {
        this.backupName = backupName;
    }

    public PluginCode getAppPluginCode() {
        return appPluginCode;
    }

    public void setAppPluginCode(PluginCode appPluginCode) {
        this.appPluginCode = appPluginCode;
    }

    public boolean isWaitForCompletion() {
        return waitForCompletion;
    }

    public void setWaitForCompletion(boolean waitForCompletion) {
        this.waitForCompletion = waitForCompletion;
    }

    public boolean isFileSystemMount() {
        return fileSystemMount;
    }

    public void setFileSystemMount(boolean fileSystemMount) {
        this.fileSystemMount = fileSystemMount;
    }

    public String getMountPath() {
        return mountPath;
    }

    public void setMountPath(String mountPath) {
        this.mountPath = mountPath;
    }

	public boolean isDataBackup() {
		return isDataBackup;
	}

	public void setDataBackup(boolean isDataBackup) {
		this.isDataBackup = isDataBackup;
	}

	public boolean isMountFromSecondary() {
		return mountFromSecondary;
	}

	public void setMountFromSecondary(boolean mountFromSecondary) {
		this.mountFromSecondary = mountFromSecondary;
	}

    public void setAppObjectId(String appObjectId) {
        this.appObjectId = appObjectId;
    }

    public String getAppObjectId() {
        return appObjectId;
    }

    public void setSecondaryLocators(List<Map<String, String>> secondaryLocators) {
        this.secondaryLocators = secondaryLocators;
    }

    public List<Map<String, String>> getSecondaryLocators() {
        return secondaryLocators;
    }

    /**
     * @return the asmRunAsName
     */
    public String getAsmRunAsName() {
        return asmRunAsName;
    }

    /**
     * @param asmRunAsName
     *            the asmRunAsName to set
     */
    public void setAsmRunAsName(String asmRunAsName) {
        this.asmRunAsName = asmRunAsName;
    }

    /**
     * @return the asmPort
     */
    public int getAsmPort() {
        return asmPort;
    }

    /**
     * @param asmPort
     *            the asmPort to set
     */
    public void setAsmPort(int asmPort) {
        this.asmPort = asmPort;
    }

}