/*
 * $Id: //depot/prod/spl/main/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/command/RemoveSmProtectResourceCommand.java#4 $
 *
 * Copyright (c) 2016 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.snapcenter.cli.command;

import static com.netapp.snapcenter.cli.command.parser.CustomStringToHostResourcesMapConverter.ORACLE_DATABASE;
import static com.netapp.snapcenter.cli.command.parser.CustomStringToHostResourcesMapConverter.RESOURCE_HOST;
import static com.netapp.snapcenter.cli.command.parser.CustomStringToHostResourcesMapConverter.RESOURCE_NAMES;
import static com.netapp.snapcenter.cli.command.parser.CustomStringToHostResourcesMapConverter.RESOURCE_NAMES_DELIMETER;
import static com.netapp.snapcenter.cli.command.parser.CustomStringToHostResourcesMapConverter.RESOURCE_TYPE;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.Parameters;
import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.smcore.plugin.contracts.SmDeleteProtectionGroupRequest;
import com.netapp.smcore.plugin.contracts.SmDeleteProtectionGroupResponse;
import com.netapp.smcore.plugin.contracts.SmDiscoverResourcesResponse;
import com.netapp.smcore.plugin.contracts.SmGetProtectionGroupsResponse;
import com.netapp.smcore.plugin.contracts.SmPluginCode;
import com.netapp.smcore.plugin.contracts.SmRequest;
import com.netapp.snapcenter.cli.command.mapper.CommandClassMapBuilder;
import com.netapp.snapcenter.cli.command.parser.CustomStringToHostResourcesMapConverter;
import com.netapp.snapcenter.cli.command.parser.NoParameterSplitter;
import com.netapp.snapcenter.cli.command.validation.DataSetCommandParametersValidationUtil;
import com.netapp.snapcenter.cli.command.validation.SmsCommandRequestValidationHelper;
import com.netapp.snapcenter.cli.contracts.PluginCode;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.helper.ResourceProtectionGroupHelper;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.ProgressBar;
import com.netapp.snapcenter.cli.utils.ProgressBarUtil;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;
import com.netapp.spl.common.restclient.config.SMSRestConfiguration;
import com.netapp.spl.common.restclient.context.SMSRestInvocationContext;

@Parameters(resourceBundle = "snapcenter-cli", commandDescriptionKey = "cli.removesmprotectresource.command.description")
public class RemoveSmProtectResourceCommand extends AbstractSMSCommand {

	private static final String COMMANDNAME = MessageBundle
			.localizeMessage(CLIMsgKey.CLI_REMOVESMPROTECTRESOURCE_COMMAND_NAME);
	private ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(CommandClassMapBuilder.class);

	@Parameter(names = { "-Resources" }, descriptionKey = "cli.removesmprotectresource.command.resources.option.description", converter = CustomStringToHostResourcesMapConverter.class, splitter = NoParameterSplitter.class, required = true)
	private List<Map<String, String>> resources = new ArrayList<Map<String, String>>();

	@Parameter(names = { "-Force" }, descriptionKey = "cli.removesmprotectresource.command.force.option.description")
	private boolean force;

	private PluginCode pluginCode;

	private List<String> protectionGroups = new ArrayList<String>();

	private ProgressBar progressBar;

	@Override
	public String getCommandName() {
		return COMMANDNAME;
	}

	@Override
	protected void processCommandArguments() throws CLIException {
		CLIUtils.promptUserToContinue(MessageBundle
				.localizeMessage(CLIMsgKey.CLI_REMOVESMPROTECTRESOURCE_PROMPT_USER_REMOVE_PROTECTION));

		// When resource is provided get the protection group name from the
		// server
		SmsCommandRequestValidationHelper smsCommandRequestValidationHelper = new SmsCommandRequestValidationHelper(
				getSnapCenterServerServiceHelper());
		try {			
			
			//validate the resources provided for SCO plugincode set
			// TODO - update this when other plugins are supported
			DataSetCommandParametersValidationUtil.validateHostResources(
					resources, PluginCode.SCO);
			
			progressBar = ProgressBarUtil
					.createAndStartProgressBar(MessageBundle
							.localizeMessage(CLIMsgKey.CLI_PROGRESSBAR_VALIDATING_RESOURCE_ALREADY_PROTECTED));
			
			for (Map<String, String> resource : resources) {

				String hostName = resource.get(RESOURCE_HOST);
				String resourceType = resource.get(RESOURCE_TYPE);
				String[] resourceNames = resource.get(RESOURCE_NAMES).split(
						RESOURCE_NAMES_DELIMETER);
				List<String> names = Arrays.asList(resourceNames);

				for (String resourceName : resourceNames) {

					setPluginCode(resourceType);

					SmDiscoverResourcesResponse smDiscoverResourcesResponse = smsCommandRequestValidationHelper
							.getDiscoverResourceDetailsFromServer(
									SmPluginCode.fromValue(pluginCode.name()),
									hostName, true);

					// find guid and validate resource
					String smObjectGuid = ResourceProtectionGroupHelper
							.getSmObjectGuidFromResourceName(
									smDiscoverResourcesResponse,
									resourceName.trim());

					// Verify whether resource is protected or not by contacting
					// CSM server. If not protected throw error that remove
					// protect will not be done.
					SmGetProtectionGroupsResponse smGetProtectionGroupsResponse = smsCommandRequestValidationHelper
							.getInternalProtectionGroupDetailsFromServer(smObjectGuid);

					if (!ResourceProtectionGroupHelper
							.isResourceAlreadyProtected(smGetProtectionGroupsResponse)) {
						throw new CLIException(
								MessageBundle
										.localizeMessage(
												CLIMsgKey.CLI_REMOVESMPROTECTRESOURCE_RESOURCE_NOT_PROTECTED,
												resourceName));
					}
					
					// Find internal protection group names for removal of protection
					String protectionGroup = ResourceProtectionGroupHelper
							.getInternalResourceGroupNameFromProtectionGroup(
									smGetProtectionGroupsResponse, smObjectGuid);
					CLIUtils.writeToLog(MessageBundle
							.localizeMessage(CLIMsgKey.CLI_REMOVESMPROTECTRESOURCE_PROTECTIONGROUP_FOUND, protectionGroup), logger, LogLevel.DEBUG);
					protectionGroups.add(protectionGroup);
				}
			}
			CLIUtils.writeToLog(MessageBundle
							.localizeMessage(CLIMsgKey.CLI_REMOVESMPROTECTRESOURCE_REMOVING_PROTECTION, protectionGroups), logger, LogLevel.DEBUG);
		} catch (Exception exception) {
			throw exception;
		} finally {
			ProgressBarUtil.stopProgressBar(progressBar);
		}
	}

	@Override
	protected <T extends SmRequest> SMSRestInvocationContext<?, ?> getCommandSpecificSMSRestInvocationContext(
			T cmdRequest) {
		SMSRestInvocationContext<SmDeleteProtectionGroupRequest, SmDeleteProtectionGroupResponse> smsRestInvocationContext = new SMSRestInvocationContext<>();
		smsRestInvocationContext
				.setRequest((SmDeleteProtectionGroupRequest) cmdRequest);
		smsRestInvocationContext
				.setResponseClass(SmDeleteProtectionGroupResponse.class);
		smsRestInvocationContext
				.setServiceUrlPath(SMSRestConfiguration.SMS_DELETE_PROTECTION_GROUP_SERVICE_PATH);
		return smsRestInvocationContext;
	}

	private void setPluginCode(String resourceType) throws CLIException {

		switch (resourceType) {
		case ORACLE_DATABASE:
			pluginCode = PluginCode.SCO;
			break;
		default:
			throw new CLIException(
					MessageBundle
							.localizeMessage(
									CLIMsgKey.CLI_HOST_RESOURCE_MAPPING_FOR_SMO_RESOURCE_TYPE_SHOULD_BE_ORACLE_DATABASE,
									resourceType));
		}

	}

	public PluginCode getPluginCode() {
		return pluginCode;
	}

	public List<String> getProtectionGroups() {
		return protectionGroups;
	}

	public void setProtectionGroups(List<String> protectionGroups) {
		this.protectionGroups = protectionGroups;
	}

	public List<Map<String, String>> getResources() {
		return resources;
	}

	public void setResources(List<Map<String, String>> resources) {
		this.resources = resources;
	}

	public boolean isForce() {
		return force;
	}

	public void setForce(boolean force) {
		this.force = force;
	}

}
