/*
 * $Id: //depot/prod/spl/main/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/command/validation/PolicyCommandParametersValidationUtil.java#1 $
 * 
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.snapcenter.cli.command.validation;

import static com.netapp.snapcenter.cli.command.parameters.VerificationPolicyParameters.DEFAULT_NUM_BACKUP_COUNT_FOR_VERIFICATION;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.lang3.StringUtils;

import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.smcore.plugin.contracts.SmPolicyType;
import com.netapp.snapcenter.cli.command.parameters.ClonePolicyParameters;
import com.netapp.snapcenter.cli.command.parameters.CommonBackupCloneVerificationPolicyParameters;
import com.netapp.snapcenter.cli.command.parameters.PolicyRetentionSettingsParameters;
import com.netapp.snapcenter.cli.command.parameters.ResourceGroupScheduleParameters;
import com.netapp.snapcenter.cli.command.parameters.SMOBackupPolicyParameters;
import com.netapp.snapcenter.cli.command.parameters.VerificationPolicyParameters;
import com.netapp.snapcenter.cli.contracts.DaysOfTheWeek;
import com.netapp.snapcenter.cli.contracts.MonthsOfTheYear;
import com.netapp.snapcenter.cli.contracts.OracleArchiveLogBackupRetentionType;
import com.netapp.snapcenter.cli.contracts.OracleBackupScope;
import com.netapp.snapcenter.cli.contracts.OracleBackupType;
import com.netapp.snapcenter.cli.contracts.OraclePruneArchiveLogDestinationType;
import com.netapp.snapcenter.cli.contracts.OraclePruneArchiveLogType;
import com.netapp.snapcenter.cli.contracts.RetentionSettingsArguments;
import com.netapp.snapcenter.cli.contracts.ScheduleType;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;

/**
 * This class contains policy command related arguments validation.
 *
 */
public class PolicyCommandParametersValidationUtil {

    private static final String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";
    private static final String PARAMETER_START_TIME = "StartDateTime";
    private static final String VALID_NAMES_REGEX = "^[a-zA-Z0-9 _-]+$";
    private static final int EXPECTED_ARGUMENT_COUNT_FOR_ORACLEBACKUPSCOPE_FULL = 2;
    private static final int SINGLE_ARGUMENT_COUNT_FOR_ORACLEBACKUPSCOPE_FULL = 1;
    private static final int EXPECTED_ARGUMENT_COUNT_FOR_ORACLEBACKUPSCOPE_DATA_OR_LOG = 1;

    private static ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(PolicyCommandParametersValidationUtil.class);

    private PolicyCommandParametersValidationUtil() {

    }

    public static void processAndValidateVerificationPolicyParameters(
            VerificationPolicyParameters verificationPolicyParameters)
            throws CLIException {
        if (verificationPolicyParameters.getNumOfBackups() < 0) {
            verificationPolicyParameters
                    .setNumOfBackups(DEFAULT_NUM_BACKUP_COUNT_FOR_VERIFICATION);
        }
    }

    public static void processAndValidateClonePolicyParameters(
            ClonePolicyParameters clonePolicyParams) throws CLIException {
        if (clonePolicyParams.getBackupPolicyName() == null) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_SMPOLICY_CLONE_POLICY_CREATION_BACKUP_POLICY_NAME_NOT_SPECIFIED));
        } else if (StringUtils.isBlank(clonePolicyParams.getBackupPolicyName())) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_SMPOLICY_CLONE_POLICY_CREATION_BACKUP_POLICY_NAME_EMPTY));
        }
    }

    public static void processAndValidateCommonBackupCloneVerificationPolicyParameters(
            ResourceGroupScheduleParameters resourceGroupScheduleParameters)
            throws CLIException {

        // Check end time is greater than the start time
        if ((resourceGroupScheduleParameters.getEndTime() != null)
                && (resourceGroupScheduleParameters.getEndTime().compareTo(
                resourceGroupScheduleParameters.getStartTime()) <= 0)) {
            throw new CLIException(MessageBundle.localizeMessage(
                    CLIMsgKey.CLI_SMPOLICY_END_TIME_LESSER_THAN_START_TIME,
                    new SimpleDateFormat(DATE_FORMAT)
                            .format(resourceGroupScheduleParameters
                                    .getStartTime())));
        }

        switch (resourceGroupScheduleParameters.getScheduleType()) {

            case NONE:
                break;

            case HOURLY:


                if (resourceGroupScheduleParameters.getOccursEveryMonth() > 0) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_OCCURS_EVERY_MONTH_INVALID));
                }

                if (CLIUtils
                        .isNotNullAndNotEmptyCollection(resourceGroupScheduleParameters
                                .getDaysOfTheMonth())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_DAYS_OF_THE_MONTH_INVALID));
                }



                if (CLIUtils
                        .isNotNullAndNotEmptyCollection(resourceGroupScheduleParameters
                                .getDaysOfTheWeek())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_DAYS_OF_THE_WEEK_INVALID));
                }

                if (CLIUtils
                        .isNotNullAndNotEmptyCollection(resourceGroupScheduleParameters
                                .getMonthsOfTheYear())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_MONTHS_OF_THE_YEAR_INVALID));
                }
                break;

            case DAILY:

                if (resourceGroupScheduleParameters.getOccursEveryMonth() > 0) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_OCCURS_EVERY_MONTH_INVALID));
                }


                if (CLIUtils
                        .isNotNullAndNotEmptyCollection(resourceGroupScheduleParameters
                                .getMonthsOfTheYear())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_MONTHS_OF_THE_YEAR_INVALID));
                }

                if (CLIUtils
                        .isNotNullAndNotEmptyCollection(resourceGroupScheduleParameters
                                .getDaysOfTheWeek())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_DAYS_OF_THE_WEEK_INVALID));
                }

                break;

            case WEEKLY:

                if (CLIUtils
                        .isNullOrEmptyCollection(resourceGroupScheduleParameters
                                .getDaysOfTheWeek())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_DAYS_OF_THE_WEEK_CANNOT_BE_EMPTY));
                }

                if (resourceGroupScheduleParameters.getOccursEveryMonth() > 0) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_OCCURS_EVERY_MONTH_INVALID));
                }

                if (CLIUtils
                        .isNotNullAndNotEmptyCollection(resourceGroupScheduleParameters
                                .getMonthsOfTheYear())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_MONTHS_OF_THE_YEAR_INVALID));
                }

                CLIUtils.removeDuplicateEntriesFromList(resourceGroupScheduleParameters
                        .getDaysOfTheWeek());
                break;

            case MONTHLY:

                if (CLIUtils
                        .isNullOrEmptyCollection(resourceGroupScheduleParameters
                                .getMonthsOfTheYear())) {
                    if (resourceGroupScheduleParameters.getOccursEveryMonth() <= 0) {
                        throw new CLIException(
                                MessageBundle
                                        .localizeMessage(CLIMsgKey.CLI_SMPOLICY_EITHER_DAYS_OF_THE_MONTH_OR_OCCURS_EVERY_MONTH_MUST_BE_SPECIFIED));
                    }
                } else {
                    if (resourceGroupScheduleParameters.getOccursEveryMonth() > 0) {
                        throw new CLIException(
                                MessageBundle
                                        .localizeMessage(CLIMsgKey.CLI_SMPOLICY_EITHER_DAYS_OF_THE_MONTH_OR_OCCURS_EVERY_MONTH_MUST_BE_SPECIFIED));
                    }

                    CLIUtils.removeDuplicateEntriesFromList(resourceGroupScheduleParameters
                            .getMonthsOfTheYear());

                }

                if (CLIUtils
                        .isNullOrEmptyCollection(resourceGroupScheduleParameters
                                .getDaysOfTheMonth())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_DAYS_OF_THE_MONTH_CANNOT_BE_EMPTY));
                }

                if (CLIUtils
                        .isNotNullAndNotEmptyCollection(resourceGroupScheduleParameters
                                .getDaysOfTheWeek())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_DAYS_OF_THE_WEEK_INVALID));
                }


                CLIUtils.removeDuplicateEntriesFromList(resourceGroupScheduleParameters
                        .getDaysOfTheMonth());
                validateDaysOfTheMonth(resourceGroupScheduleParameters
                        .getDaysOfTheMonth());
                break;

            default:
                break;
        }
    }


    private static void validateDaysOfTheMonth(List<Integer> daysOfTheMonth)
            throws CLIException {
        for (Integer day : daysOfTheMonth) {
            if (day.intValue() < 1 || day.intValue() > 31) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(
                                        CLIMsgKey.CLI_SMPOLICY_DAYS_OF_THE_MONTH_INVALID_VALUES,
                                        day.intValue()));
            }
        }
    }

    public static void processAndValidateSMOBackupPolicyParameters(
            SMOBackupPolicyParameters smoBackupPolicyParams)
            throws CLIException {

        if (smoBackupPolicyParams.getOracleBackupType() == null) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ORACLE_BACKUP_TYPE_NOT_SPECIFIED));
        }
        if (smoBackupPolicyParams.getOracleBackupScope() == null) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ORACLE_BACKUP_SCOPE_NOT_SPECIFIED));
        }

        if (smoBackupPolicyParams.getOracleBackupType().equals(
                OracleBackupType.OFFLINESHUTDOWN)
                || smoBackupPolicyParams.getOracleBackupType().equals(
                OracleBackupType.OFFLINEMOUNT)) {
            if (!smoBackupPolicyParams.getOracleBackupScope().equals(
                    OracleBackupScope.DATA)) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(
                                        CLIMsgKey.CLI_SMPOLICY_INVALID_ORACLE_BACKUP_SCOPE_AND_TYPE,
                                        smoBackupPolicyParams
                                                .getOracleBackupScope().name(),
                                        smoBackupPolicyParams
                                                .getOracleBackupType().name()));
            }
        }

        if (smoBackupPolicyParams.isVerificationAfterBackup()) {
            if (smoBackupPolicyParams.getVerificationPolicy() == null) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_SMPOLICY_BACKUP_POLICY_CREATION_VERIFICATION_POLICY_NAME_NOT_SPECIFIED));
            }
            if (StringUtils.isBlank(smoBackupPolicyParams
                    .getVerificationPolicy())) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_SMPOLICY_BACKUP_POLICY_CREATION_VERIFICATION_POLICY_NAME_EMPTY));
            }
            if (smoBackupPolicyParams.getOracleBackupScope().equals(
                    OracleBackupScope.LOG)) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_SMPOLICY_VERIFICATION_NOT_VALID_BACKUP_SCOPE));
            }
        }

        // validate archive log pruning related parameters
        if (smoBackupPolicyParams.isOraclePruneArchiveLog()) {
            if (smoBackupPolicyParams.getOracleBackupScope().equals(
                    OracleBackupScope.DATA)) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ARCHIVELOG_PRUNING_NOT_VALID_BACKUP_SCOPE));
            } else if (smoBackupPolicyParams.getOraclePruneArchiveLogType() == null) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ORACLE_PRUNE_ARCHIVE_LOG_TYPE_NOT_SPECIFIED));
            } else if (smoBackupPolicyParams.getOraclePruneArchiveLogType()
                    .equals(OraclePruneArchiveLogType.OlderLogs)) {
                if (smoBackupPolicyParams
                        .getOraclePruneArchiveLogOlderThanDays() == null) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ORACLE_PRUNE_ARCHIVE_LOG_OLDER_THAN_DAYS_NOT_SPECIFIED));
                } else if (smoBackupPolicyParams
                        .getOraclePruneArchiveLogOlderThanDays().isInfinite()
                        || smoBackupPolicyParams
                        .getOraclePruneArchiveLogOlderThanDays()
                        .isNaN()
                        || smoBackupPolicyParams
                        .getOraclePruneArchiveLogOlderThanDays()
                        .doubleValue() < 0) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_SMPOLICY_ORACLE_PRUNE_ARCHIVE_LOG_OLDER_THAN_DAYS_INVALID_VALUES,
                                            smoBackupPolicyParams
                                                    .getOraclePruneArchiveLogOlderThanDays()));
                }
            }

            // all validations done for pruning related parameters,
            // set the prune destination type default value if it is not set
            // we cannot set it as part of class parameter declaration as there
            // are validations based on it
            // being set or not by the user
            if (smoBackupPolicyParams.getOraclePruneArchiveLogDestinationType() == null) {
                smoBackupPolicyParams
                        .setOraclePruneArchiveLogDestinationType(OraclePruneArchiveLogDestinationType.AllDestinations);
            }
        }

        if (!smoBackupPolicyParams.isOraclePruneArchiveLog()
                && (smoBackupPolicyParams.getOraclePruneArchiveLogType() != null
                || smoBackupPolicyParams
                .getOraclePruneArchiveLogOlderThanDays() != null || smoBackupPolicyParams
                .getOraclePruneArchiveLogDestinationType() != null)) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ORACLE_PRUNE_ARCHIVE_LOG_NOT_SPECIFIED));
        }

        // validate Log retention related parameters
        if (smoBackupPolicyParams.isOracleDeleteArchiveLogBackup()) {
            if (smoBackupPolicyParams.getOracleArchiveLogBackupRetentionType() == null) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ORACLE_ARCHIVE_LOG_BACKUP_RETENTION_TYPE_NOT_SPECIFIED));
            } else if (smoBackupPolicyParams
                    .getOracleArchiveLogBackupRetentionType().equals(
                            OracleArchiveLogBackupRetentionType.CountBase)) {
                if (smoBackupPolicyParams
                        .getOracleDeleteArchiveLogBackupCounts() == null) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ORACLE_DELETE_ARCHIVE_LOG_BACKUP_COUNTS_NOT_SPECIFIED));
                } else if (smoBackupPolicyParams
                        .getOracleDeleteArchiveLogBackupCounts().intValue() < 0) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_SMPOLICY_ORACLE_DELETE_ARCHIVE_LOG_BACKUP_COUNTS_INVALID_VALUES,
                                            smoBackupPolicyParams
                                                    .getOracleDeleteArchiveLogBackupCounts()));
                }

            } else if (smoBackupPolicyParams
                    .getOracleArchiveLogBackupRetentionType().equals(
                            OracleArchiveLogBackupRetentionType.DayBase)) {
                if (smoBackupPolicyParams.getOracleDeleteArchiveLogBackupDays() == null) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ORACLE_DELETE_ARCHIVE_LOG_BACKUP_DAYS_NOT_SPECIFIED));
                } else if (smoBackupPolicyParams
                        .getOracleDeleteArchiveLogBackupDays().isInfinite()
                        || smoBackupPolicyParams
                        .getOracleDeleteArchiveLogBackupDays().isNaN()
                        || smoBackupPolicyParams
                        .getOracleDeleteArchiveLogBackupDays()
                        .doubleValue() < 0) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_SMPOLICY_ORACLE_DELETE_ARCHIVE_LOG_BACKUP_DAYS_INVALID_VALUES,
                                            smoBackupPolicyParams
                                                    .getOracleDeleteArchiveLogBackupDays()));
                }

            }
        }

        if (!smoBackupPolicyParams.isOracleDeleteArchiveLogBackup()
                && (smoBackupPolicyParams
                .getOracleArchiveLogBackupRetentionType() != null
                || smoBackupPolicyParams
                .getOracleDeleteArchiveLogBackupCounts() != null || smoBackupPolicyParams
                .getOracleDeleteArchiveLogBackupDays() != null)) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_SMPOLICY_ORACLE_DELETE_ARCHIVE_LOG_BACKUP_NOT_SPECIFIED));
        }
    }

    public static void isValidPolicyName(String policyName) throws CLIException {
        if (!policyName.matches(VALID_NAMES_REGEX)) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_INVALID_POLICY_NAME));
        }
    }

    /**
     * Validates and process the arguments specified in the option
     * -RetentionSettings
     *
     * @param retentionSettings
     * @param smoBackupPolicyParameters
     * @param commonBackupCloneVerificationPolicyParameters
     * @throws CLIException
     */
    public static List<PolicyRetentionSettingsParameters> processPolicyRetentionParameters(
            List<Map<String, String>> retentionSettings,
            SMOBackupPolicyParameters smoBackupPolicyParameters,
            CommonBackupCloneVerificationPolicyParameters commonBackupCloneVerificationPolicyParameters)
            throws CLIException {

        CLIUtils.writeToLog("Processing  PolicyRetentionParameters "
                + retentionSettings, logger, LogLevel.DEBUG);
        List<PolicyRetentionSettingsParameters> listPolicyRetentionSettingsParametersMappers = new ArrayList<PolicyRetentionSettingsParameters>();

        List<ScheduleType> scheduleTypesOfOptionScheduleType = commonBackupCloneVerificationPolicyParameters
                .getScheduleTypeAsListOfScheduleType();
        OracleBackupScope oracleBackupScopeOfOptionOracleBackupScope = smoBackupPolicyParameters
                .getOracleBackupScope();

        /*
         * If retention settings are not provided as part of retention settings
         * option then default values will be filled.
         */
        Map<ScheduleType, OracleBackupScope> mapOfDefaultValuesToBeFilled = new HashMap<ScheduleType, OracleBackupScope>();

        if (!retentionSettings.isEmpty()) {
            /*
             * When retentionsettings details are not provided for all the
             * schedule types specified in the option '-ScheduleType', those
             * will be filled with default values. For eg: -scheduletype
             * 'hourly,weekly' -retentionsettings
             * {'scheduleType=weekly,backupscope=data,retentiondays=1'}
             * retentionsettings has details for only weekly not hourly, for
             * hourly default values will be filled.
             */

            Map<ScheduleType, List<PolicyRetentionSettingsParameters>> mapOfScheduleTypeAndRetentionSettingsParameters = validateAndGetPolicyRetentionSettingsParametersMapperFromRetentionSettings(
                    retentionSettings,
                    oracleBackupScopeOfOptionOracleBackupScope);

            /*
             * It validates the schedule type specified in -ScheduleType with
             * one of the -RetentionSetting scheduleType parameter, if other
             * than the schedule type specified in -ScheduleType it will throw
             * an exception.
             */
            if (!scheduleTypesOfOptionScheduleType
                    .containsAll(mapOfScheduleTypeAndRetentionSettingsParameters
                            .keySet())) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_SMPOLICY_MISSMATCH_SCHEDULE_TYPE_RETENTION_SETTINGS_VALUES));

            }

            // Add the missing schedule types from retention settings option to
            // map, to fill default values.
            for (ScheduleType scheduleType : scheduleTypesOfOptionScheduleType) {
                if (!mapOfScheduleTypeAndRetentionSettingsParameters
                        .containsKey(scheduleType)) {
                    mapOfDefaultValuesToBeFilled.put(scheduleType,
                            oracleBackupScopeOfOptionOracleBackupScope);
                }

            }

            /*
             * Check for duplicate arguments specified in retention settings
             * option, and add the arguments to the list of
             * PolicyRetentionSettingsParametersMapper.
             */
            for (Entry<ScheduleType, List<PolicyRetentionSettingsParameters>> mapEntryOfScheduleTypeAndRetentionSettingsParameters : mapOfScheduleTypeAndRetentionSettingsParameters
                    .entrySet()) {
                switch (oracleBackupScopeOfOptionOracleBackupScope) {

                    case FULL:
                        if (mapEntryOfScheduleTypeAndRetentionSettingsParameters
                                .getValue().size() > EXPECTED_ARGUMENT_COUNT_FOR_ORACLEBACKUPSCOPE_FULL) {
                            throw new CLIException(
                                    MessageBundle
                                            .localizeMessage(CLIMsgKey.CLI_SMPOLICY_DUPLICATE_RETENTION_SETTINGS_ARGUMENTS));
                        }
                        if (mapEntryOfScheduleTypeAndRetentionSettingsParameters
                                .getValue().size() == SINGLE_ARGUMENT_COUNT_FOR_ORACLEBACKUPSCOPE_FULL) {
                            List<PolicyRetentionSettingsParameters> listOfPolicyRetentionSettingsParametersMapper = mapEntryOfScheduleTypeAndRetentionSettingsParameters
                                    .getValue();
                            for (PolicyRetentionSettingsParameters policyRetentionSettingsParametersMapper : listOfPolicyRetentionSettingsParametersMapper) {
                                if (policyRetentionSettingsParametersMapper
                                        .getOracleBackupScope().equals(
                                                OracleBackupScope.DATA)) {
                                    mapOfDefaultValuesToBeFilled.put(
                                            mapEntryOfScheduleTypeAndRetentionSettingsParameters
                                                    .getKey(),
                                            OracleBackupScope.LOG);
                                } else {
                                    mapOfDefaultValuesToBeFilled.put(
                                            mapEntryOfScheduleTypeAndRetentionSettingsParameters
                                                    .getKey(),
                                            OracleBackupScope.DATA);
                                }
                            }
                        }
                        break;

                    case DATA:
                        // Common validation for both DATA and LOG

                    case LOG:
                        if (mapEntryOfScheduleTypeAndRetentionSettingsParameters
                                .getValue().size() > EXPECTED_ARGUMENT_COUNT_FOR_ORACLEBACKUPSCOPE_DATA_OR_LOG) {
                            throw new CLIException(
                                    MessageBundle
                                            .localizeMessage(CLIMsgKey.CLI_SMPOLICY_DUPLICATE_RETENTION_SETTINGS_ARGUMENTS));
                        }
                        break;

                }
                listPolicyRetentionSettingsParametersMappers
                        .addAll(mapEntryOfScheduleTypeAndRetentionSettingsParameters
                                .getValue());
            }

        } else {
            /*
             * When retentionsettings details are not provided for none of the
             * schedule types specified in the option '-ScheduleType', those
             * will be filled with default values
             */

            for (ScheduleType scheduleType : scheduleTypesOfOptionScheduleType) {
                mapOfDefaultValuesToBeFilled.put(scheduleType,
                        oracleBackupScopeOfOptionOracleBackupScope);
            }

        }

        addDefaultValues(listPolicyRetentionSettingsParametersMappers,
                mapOfDefaultValuesToBeFilled);

        return listPolicyRetentionSettingsParametersMappers;
    }

    /*
     * It fills the default values for each of the schedule type passed.
     */
    private static void addDefaultValues(
            List<PolicyRetentionSettingsParameters> listPolicyRetentionSettingsParametersMappers,
            Map<ScheduleType, OracleBackupScope> mapOfDefaultValuesToBeFilled) {

        for (Entry<ScheduleType, OracleBackupScope> mapEntryOfDefaultValuesToBeFilled : mapOfDefaultValuesToBeFilled
                .entrySet()) {
            ScheduleType scheduleType = mapEntryOfDefaultValuesToBeFilled
                    .getKey();
            OracleBackupScope oracleBackupScope = mapEntryOfDefaultValuesToBeFilled
                    .getValue();
            PolicyRetentionSettingsParameters parametersMapper = new PolicyRetentionSettingsParameters();
            // By default retention will be set to retention count.
            parametersMapper
                    .setDeleteBackupInExcess(PolicyRetentionSettingsParameters.NUMBER_OF_SNAPSHOTS_TO_KEEP);
            parametersMapper.setScheduleType(scheduleType);
            CLIUtils.writeToLog("adding default retention settings values for "
                    + scheduleType, logger, LogLevel.DEBUG);

            switch (oracleBackupScope) {

                case FULL:

                    PolicyRetentionSettingsParameters parametersMapper1 = new PolicyRetentionSettingsParameters();
                    parametersMapper1
                            .setOracleBackupScope(OracleBackupScope.DATA);
                    parametersMapper1.setScheduleType(scheduleType);
                    parametersMapper1
                            .setDeleteBackupInExcess(PolicyRetentionSettingsParameters.NUMBER_OF_SNAPSHOTS_TO_KEEP);
                    listPolicyRetentionSettingsParametersMappers
                            .add(parametersMapper1);

                    PolicyRetentionSettingsParameters parametersMapper2 = new PolicyRetentionSettingsParameters();
                    parametersMapper2
                            .setOracleBackupScope(OracleBackupScope.LOG);
                    parametersMapper2.setScheduleType(scheduleType);
                    parametersMapper2
                            .setDeleteBackupInExcess(PolicyRetentionSettingsParameters.NUMBER_OF_SNAPSHOTS_TO_KEEP);
                    listPolicyRetentionSettingsParametersMappers
                            .add(parametersMapper2);
                    break;

                case DATA:
                    parametersMapper.setOracleBackupScope(oracleBackupScope);
                    listPolicyRetentionSettingsParametersMappers
                            .add(parametersMapper);
                    break;

                case LOG:
                    parametersMapper.setOracleBackupScope(oracleBackupScope);
                    listPolicyRetentionSettingsParametersMappers
                            .add(parametersMapper);
                    break;

            }
        }

    }

    /*
     * This function performs validation of the retention settings parameters
     * and returns the map of ScheduleType and list of
     * PolicyRetentionSettingsParametersMapper, where schedule type is extracted
     * from retentionsettings option.
     */
    private static Map<ScheduleType, List<PolicyRetentionSettingsParameters>> validateAndGetPolicyRetentionSettingsParametersMapperFromRetentionSettings(
            List<Map<String, String>> retentionSettings,
            OracleBackupScope oracleBackupScopeOfOptionOracleBackupScope)
            throws CLIException {

        Map<ScheduleType, List<PolicyRetentionSettingsParameters>> mapOfScheduleTypeAndRetentionSettingsParameters = new HashMap<ScheduleType, List<PolicyRetentionSettingsParameters>>();
        for (Map<String, String> retentionSetting : retentionSettings) {

            ScheduleType scheduleTypeFromOptionRetentionSettings;
            // SChedule type is not specified in retentionsettings option it
            // will assigned to none(On-Demand backup policy)
            if (StringUtils.isNotBlank(retentionSetting
                    .get(RetentionSettingsArguments.SCHEDULETYPE.value()))) {
                scheduleTypeFromOptionRetentionSettings = ScheduleType
                        .fromValue(retentionSetting
                                .get(RetentionSettingsArguments.SCHEDULETYPE
                                        .value()));
            } else {
                scheduleTypeFromOptionRetentionSettings = ScheduleType.NONE;
            }

            OracleBackupScope backupScopeFromOptionRetentionSettings = OracleBackupScope
                    .fromValue(retentionSetting
                            .get(RetentionSettingsArguments.BACKUPSCOPE.value()));

            /*
             * Value provided in -OracleBackupScope and backuptype of retention
             * settings should match. For ex: if -OracleBackupScope = LOG then
             * backuptype of retention settings should not be DATA. For
             * -OracleBackupScope = FULL, backuptype of retention settings can
             * be both/any one of them but cannot be FULL coz SmBackupType takes
             * only DATA, LOG, LOG_SNAPSHOT
             */
            switch (oracleBackupScopeOfOptionOracleBackupScope) {

                case DATA:
                    if (!backupScopeFromOptionRetentionSettings
                            .equals(OracleBackupScope.DATA)) {
                        throw new CLIException(
                                MessageBundle
                                        .localizeMessage(CLIMsgKey.CLI_SMPOLICY_MISSMATCH_ORACLE_BACKUP_SCOPE_RETENTION_SETTINGS_VALUES));
                    }
                    break;

                case LOG:
                    if (!backupScopeFromOptionRetentionSettings
                            .equals(OracleBackupScope.LOG)) {
                        throw new CLIException(
                                MessageBundle
                                        .localizeMessage(CLIMsgKey.CLI_SMPOLICY_MISSMATCH_ORACLE_BACKUP_SCOPE_RETENTION_SETTINGS_VALUES));
                    }
                    break;

                case FULL:
                    if (backupScopeFromOptionRetentionSettings
                            .equals(OracleBackupScope.FULL)) {
                        throw new CLIException(
                                MessageBundle
                                        .localizeMessage(
                                                CLIMsgKey.CLI_SMPOLICY_INVALID_ORACLE_BACKUP_SCOPE_RETENTION_SETTINGS_VALUES,
                                                backupScopeFromOptionRetentionSettings));
                    }
                    break;

            }

            // Both RETENTIONCOUNT and RETENTIONDAYS cannot be specified at the
            // same time.
            if ((retentionSetting.get(RetentionSettingsArguments.RETENTIONCOUNT
                    .value()) != null)
                    && (retentionSetting
                    .get(RetentionSettingsArguments.RETENTIONDAYS
                            .value()) != null)) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_SMPOLICY_WRONG_RETENTION_SETTINGS_ARGUMENTS));

            }

            /*
             * if map mapOfScheduleTypeAndRetentionSettingsParameters doesn't
             * contains the key, create new list of
             * PolicyRetentionSettingsParametersMapper and add it to map.
             */
            if (!mapOfScheduleTypeAndRetentionSettingsParameters
                    .containsKey(scheduleTypeFromOptionRetentionSettings)) {
                mapOfScheduleTypeAndRetentionSettingsParameters.put(
                        scheduleTypeFromOptionRetentionSettings,
                        new ArrayList<PolicyRetentionSettingsParameters>());
            }

            List<PolicyRetentionSettingsParameters> listOfPolicyRetentionSettingsParametersMappers = mapOfScheduleTypeAndRetentionSettingsParameters
                    .get(scheduleTypeFromOptionRetentionSettings);

            PolicyRetentionSettingsParameters policyRetentionSettingsParametersMapper = new PolicyRetentionSettingsParameters();
            policyRetentionSettingsParametersMapper
                    .setScheduleType(scheduleTypeFromOptionRetentionSettings);
            policyRetentionSettingsParametersMapper
                    .setOracleBackupScope(backupScopeFromOptionRetentionSettings);

            if (retentionSetting.get(RetentionSettingsArguments.RETENTIONCOUNT
                    .value()) != null) {
                policyRetentionSettingsParametersMapper
                        .setDeleteBackupInExcess(Integer.parseInt(retentionSetting
                                .get(RetentionSettingsArguments.RETENTIONCOUNT
                                        .value())));
            } else {
                policyRetentionSettingsParametersMapper
                        .setDeleteBackupOlderThan(Integer.parseInt(retentionSetting
                                .get(RetentionSettingsArguments.RETENTIONDAYS
                                        .value())));
            }

            listOfPolicyRetentionSettingsParametersMappers
                    .add(policyRetentionSettingsParametersMapper);
        }
        return mapOfScheduleTypeAndRetentionSettingsParameters;

    }

    /*
     * It validates the schedule type for negative case, where schedules are
     * provided for backup policy
     */
    public static void validateIfScheduleIsValid(
            CommonBackupCloneVerificationPolicyParameters commonParams,
            List<PolicyRetentionSettingsParameters> listPolicyRetentionSettingsParameters,
            String policyTypeString) throws CLIException {

        SmPolicyType policyType = SmPolicyType.valueOf(policyTypeString);
        List<ScheduleType> scheduleTypes = commonParams
                .getScheduleTypeAsListOfScheduleType();

        /*
         * It tests for negative case, since for verification policy still it
         * takes the schedule details. TODO: after verification changes in
         * server, it has to be removed.
         */
        if (policyType.equals(SmPolicyType.BACKUP)) {
            // TODO: Delete below code
            /*
             * if ((commonParams.getDaysOfTheMonth() != null && commonParams
             * .getDaysOfTheMonth().size() > 0) ||
             * (commonParams.getDaysOfTheWeekAsStringList() != null &&
             * commonParams .getDaysOfTheMonth().size() > 0) ||
             * commonParams.getEndTime() != null ||
             * (commonParams.getMonthsOfTheYearAsStringList() != null &&
             * commonParams .getDaysOfTheMonth().size() > 0) ||
             * commonParams.getOccursEveryMinute() != 0 ||
             * commonParams.getOccursEveryMonth() != 0 ||
             * commonParams.getStartTime() != null) {
             * 
             * throw new CLIException( MessageBundle .localizeMessage(CLIMsgKey.
             * CLI_ADDSMPOLICY_COMMAND_BACKUP_SCHEDULE_OPTION_DESCRIPTION)); }
             */

            if (!CLIUtils.isEmpty(commonParams
                    .getListVerificationScheduleTypes())
                    && !(commonParams.getScheduleTypeAsListOfScheduleType()
                    .containsAll(commonParams
                            .getListVerificationScheduleTypes()))) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_VERIFICATION_SCHEDULE_TYPES_NOT_MATCHING_WITH_SPECIFIED_BACKUP_SCHEDULE_TYPES));
            }
        }
    }

}
