/*
 * $Id: //depot/prod/spl/main/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/command/validation/RestoreCommandParametersValidationUtil.java#1 $
 * 
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.snapcenter.cli.command.validation;

import static com.netapp.snapcenter.cli.command.parameters.SMORestoreParameters.PARAMETER_OPEN_PLUGGABLE_DATABASE_AFTER_RECOVERY;
import static com.netapp.snapcenter.cli.command.parameters.SMORestoreParameters.PARAMETER_OPEN_TABLESPACE_AFTER_RECOVERY;
import static com.netapp.snapcenter.cli.command.parameters.SMORestoreParameters.PARAMETER_ORACLE_FULL_RESTORE;
import static com.netapp.snapcenter.cli.command.parameters.SMORestoreParameters.PARAMETER_ORACLE_PLUGGABLE_DATABASES;
import static com.netapp.snapcenter.cli.command.parameters.SMORestoreParameters.PARAMETER_ORACLE_TABLESPACES;

import org.apache.commons.lang3.StringUtils;

import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.smcore.plugin.contracts.SmGetHostResponse;
import com.netapp.snapcenter.cli.command.parameters.SDURestoreParameters;
import com.netapp.snapcenter.cli.command.parameters.SMORestoreParameters;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.helper.SnapCenterServerServiceHelper;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;

/**
 * This class contains restore command related arguments validation.
 *
 */
public class RestoreCommandParametersValidationUtil {

    private static final String VALID_NAMES_REGEX = "^[a-zA-Z0-9][a-zA-Z0-9_]*";
    private static ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(RestoreCommandParametersValidationUtil.class);

    private RestoreCommandParametersValidationUtil() {

    }
    
    public static void processAndValidateSDURestoreParameters(
            SDURestoreParameters sduRestoreParams) throws CLIException {
        if (sduRestoreParams.isForceInplaceRestore()) {
            CLIUtils.writeToLogAndConsole(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_SDU_INPLACE_RESTORE_SELECTED),
                    logger, LogLevel.WARN);
        }
    }

    public static void processAndValidateSMORestoreParameters(
            SMORestoreParameters smoRestoreParams, SnapCenterServerServiceHelper snapCenterServerServiceHelper) throws CLIException {

    	validateTargetHostForRestore(smoRestoreParams, snapCenterServerServiceHelper);
    	
        validatePdbAndTablespaceNames(smoRestoreParams);

        // Validation for partial restore parameters
        if (!smoRestoreParams.isOracleFullRestore()) {
            validateForPartialRestoreOptions(smoRestoreParams);
        }

        // Validation for recovery parameters
        if (!smoRestoreParams.isOracleSkipRecovery()) {
            validateForRecoveryOptions(smoRestoreParams);
        } else if (smoRestoreParams.getOracleUntilScn() != null
                || smoRestoreParams.getOracleUntilTime() != null) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_INVALID_RECOVERY_OPTIONS));
        }

        validateAndSetOracleOpenDatabaseAfterRecovery(smoRestoreParams);

        if (smoRestoreParams.isOracleFullRestore()) {
            if (CLIUtils.isNotNullAndNotEmptyCollection(smoRestoreParams
                    .getOracleTablespaces())) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(
                                        CLIMsgKey.CLI_PARAMETER_CANNOT_BE_USED_WHEN_ANOTHER_PARAMETER_SET,
                                        PARAMETER_ORACLE_TABLESPACES,
                                        PARAMETER_ORACLE_FULL_RESTORE));
            } else if (CLIUtils.isNotNullAndNotEmptyCollection(smoRestoreParams
                    .getOraclePluggableDatabases())) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(
                                        CLIMsgKey.CLI_PARAMETER_CANNOT_BE_USED_WHEN_ANOTHER_PARAMETER_SET,
                                        PARAMETER_ORACLE_PLUGGABLE_DATABASES,
                                        PARAMETER_ORACLE_FULL_RESTORE));
            }
        }

        if (smoRestoreParams.isOracleControlFileRestore()) {
            String controlFileRestoreSelected = MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_CONTROL_FILE_RESTORE_OPTION_IS_CHOSEN);
            CLIUtils.writeToLogAndConsole(controlFileRestoreSelected, logger,
                    LogLevel.INFO);
        }

        if (smoRestoreParams.isOracleRedologFileRestore()) {
            String redologFileRestoreSelected = MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_REDOLOG_FILE_RESTORE_OPTION_IS_CHOSEN);
            CLIUtils.writeToLogAndConsole(redologFileRestoreSelected, logger,
                    LogLevel.INFO);
        }

        if (smoRestoreParams.isOracleSkipRecovery()) {
            String skipRecoverySelected = MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_SKIP_RECOVERY_OPTION_IS_CHOSEN);
            CLIUtils.writeToLogAndConsole(skipRecoverySelected, logger,
                    LogLevel.INFO);
        } else {
            if (smoRestoreParams.isOracleFullRestore()
                    && (smoRestoreParams.getOracleUntilScn() == null && smoRestoreParams
                            .getOracleUntilTime() == null)) {
                String allLogsSelected = MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_ALL_LOGS_RECOVERY_WILL_BE_PERFORMED);
                CLIUtils.writeToLogAndConsole(allLogsSelected, logger,
                        LogLevel.INFO);
            }
        }

    }

	private static void validateTargetHostForRestore(
			SMORestoreParameters smoRestoreParams,
			SnapCenterServerServiceHelper snapCenterServerServiceHelper)
			throws CLIException {
		String hostName = smoRestoreParams.getRestoreOnClusterHost();
		if (StringUtils.isNotBlank(hostName)) {
			SmsCommandRequestValidationHelper smsCommandRequestValidationHelper = new SmsCommandRequestValidationHelper(
					snapCenterServerServiceHelper);
			SmGetHostResponse smGetHostResponse = smsCommandRequestValidationHelper
					.getSmHost(hostName);
			if (smGetHostResponse.getHostInfo().getHosts().getSmHost()
					.isEmpty()) {
				String error = MessageBundle.localizeMessage(
						CLIMsgKey.CLI_PARAMETER_SC_OPERATION_INVALID_HOST,
						hostName);
				throw new CLIException(error);
			}
		}
	}
    
    private static void validateForPartialRestoreOptions(
            SMORestoreParameters smoRestoreParams) throws CLIException {
        if (CLIUtils.isNullOrEmptyCollection(smoRestoreParams
                .getOracleTablespaces())
                && CLIUtils.isNullOrEmptyCollection(smoRestoreParams
                        .getOraclePluggableDatabases())
                && !smoRestoreParams.isOracleControlFileRestore()
                && !smoRestoreParams.isOracleRedologFileRestore()) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_NOTHING_SPECIFIED_TO_RESTORE));
        } else if (CLIUtils.isNotNullAndNotEmptyCollection(smoRestoreParams
                .getOracleTablespaces())
                && CLIUtils.isNotNullAndNotEmptyCollection(smoRestoreParams
                        .getOraclePluggableDatabases())) {
            /*
             * Only Pdbs or tablespaces or tablespaces of single pdb can be
             * restored. Tablespaces of multiple pdbs cannot be performed.
             */
            if (smoRestoreParams.getOraclePluggableDatabases().size() != 1) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_SPECIFY_EITHER_TABLESPACES_OR_PDBS_OR_TABLESPACES_OF_SINGLE_PDB));
            }
        }

    }

    private static void validateForRecoveryOptions(
            SMORestoreParameters smoRestoreParams) throws CLIException {

        if (!smoRestoreParams.isOracleFullRestore()
                && (smoRestoreParams.getOracleUntilScn() != null || smoRestoreParams
                        .getOracleUntilTime() != null)) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_INCOMPLETE_RECOVERY_WITH_PARTIAL_RESTORE));
        }

        if (smoRestoreParams.getOracleUntilScn() != null
                && smoRestoreParams.getOracleUntilTime() != null) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_NOT_A_VALID_CONFIGURATION_FOR_RECOVERY));
        }

        for (String logsPath : smoRestoreParams.getAlternateArchiveLogPaths()) {
            if (StringUtils.isBlank(logsPath)) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_ORACLE_ALTERNATE_ARCHIVE_LOG_PATH_CANNOT_BE_EMPTY));
            }
        }

    }

    private static void validatePdbAndTablespaceNames(
            SMORestoreParameters smoRestoreParams) throws CLIException {
        if (!smoRestoreParams.getOraclePluggableDatabases().isEmpty()) {
            for (String oraclePluggableDatabase : smoRestoreParams
                    .getOraclePluggableDatabases()) {
                if (!oraclePluggableDatabase.matches(VALID_NAMES_REGEX)) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_PARAMETER_RESTORESMBACKUP_INVALID_VALUE_FOR_PARAMETER,
                                            oraclePluggableDatabase,
                                            PARAMETER_ORACLE_PLUGGABLE_DATABASES));
                }
            }
        }
        if (!smoRestoreParams.getOracleTablespaces().isEmpty()) {
            for (String oracleTablespace : smoRestoreParams
                    .getOracleTablespaces()) {
                if (!oracleTablespace.matches(VALID_NAMES_REGEX)) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_PARAMETER_RESTORESMBACKUP_INVALID_VALUE_FOR_PARAMETER,
                                            oracleTablespace,
                                            PARAMETER_ORACLE_TABLESPACES));
                }
            }
        }
    }

    private static void validateAndSetOracleOpenDatabaseAfterRecovery(
            SMORestoreParameters smoRestoreParams) throws CLIException {

        boolean oracleSkipRecovery = smoRestoreParams.isOracleSkipRecovery();
        boolean oracleOpenDatabaseAfterRecovery = smoRestoreParams
                .isOracleOpenDatabaseAfterRecovery();
        boolean oracleOpenPluggableDatabaseAfterRecovery = smoRestoreParams
                .isOracleOpenPluggableDatabaseAfterRecovery();
        boolean oracleOpenTablespaceAfterRecovery = smoRestoreParams
                .isOracleOpenTablespaceAfterRecovery();
        if (oracleSkipRecovery) {
            if (oracleOpenDatabaseAfterRecovery
                    || oracleOpenPluggableDatabaseAfterRecovery
                    || oracleOpenTablespaceAfterRecovery) {
                throw new CLIException(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_PARAMETER_RESTORESMBACKUP_OPENDATABASE_WHEN_SKIPRECOVERY));
            }
        } else {
            if (oracleOpenDatabaseAfterRecovery) {
                if (oracleOpenPluggableDatabaseAfterRecovery
                        || oracleOpenTablespaceAfterRecovery) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_PARAMETER_RESTORESMBACKUP_SINGLE_OPENDATABASE_PARAMETER));
                }
                if (CLIUtils.isNotNullAndNotEmptyCollection(smoRestoreParams
                        .getOracleTablespaces())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_PARAMETER_WHEN_ANOTHER_PARAMETER_USED,
                                            PARAMETER_OPEN_TABLESPACE_AFTER_RECOVERY,
                                            PARAMETER_ORACLE_TABLESPACES));
                } else if (CLIUtils
                        .isNotNullAndNotEmptyCollection(smoRestoreParams
                                .getOraclePluggableDatabases())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_PARAMETER_WHEN_ANOTHER_PARAMETER_USED,
                                            PARAMETER_OPEN_PLUGGABLE_DATABASE_AFTER_RECOVERY,
                                            PARAMETER_ORACLE_PLUGGABLE_DATABASES));
                }
            } else if (oracleOpenPluggableDatabaseAfterRecovery) {
                if (oracleOpenTablespaceAfterRecovery) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(CLIMsgKey.CLI_PARAMETER_RESTORESMBACKUP_SINGLE_OPENDATABASE_PARAMETER));
                } else {
                    if (CLIUtils
                            .isNotNullAndNotEmptyCollection(smoRestoreParams
                                    .getOracleTablespaces())) {
                        throw new CLIException(
                                MessageBundle
                                        .localizeMessage(
                                                CLIMsgKey.CLI_PARAMETER_WHEN_ANOTHER_PARAMETER_USED,
                                                PARAMETER_OPEN_TABLESPACE_AFTER_RECOVERY,
                                                PARAMETER_ORACLE_TABLESPACES));
                    } else if (CLIUtils
                            .isNullOrEmptyCollection(smoRestoreParams
                                    .getOraclePluggableDatabases())) {
                        throw new CLIException(
                                MessageBundle
                                        .localizeMessage(
                                                CLIMsgKey.CLI_PARAMETER_CANNOT_BE_EMPTY_WHEN_ANOTHER_PARAMETER_SET,
                                                PARAMETER_ORACLE_PLUGGABLE_DATABASES,
                                                PARAMETER_OPEN_PLUGGABLE_DATABASE_AFTER_RECOVERY));
                    }
                }
            } else if (oracleOpenTablespaceAfterRecovery) {
                if (CLIUtils.isNullOrEmptyCollection(smoRestoreParams
                        .getOracleTablespaces())) {
                    throw new CLIException(
                            MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_PARAMETER_CANNOT_BE_EMPTY_WHEN_ANOTHER_PARAMETER_SET,
                                            PARAMETER_ORACLE_TABLESPACES,
                                            PARAMETER_OPEN_TABLESPACE_AFTER_RECOVERY));
                }
            }
        }
    }
}
