/*
 * $Id: //depot/prod/spl/main/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/helper/SnapCenterServerServiceFactory.java#1 $
 * 
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.snapcenter.cli.helper;

import static com.netapp.spl.common.restclient.helper.PropertiesFileReader.AGENT_CONFIG_SNAPCENTER_SERVER_HOST;
import static com.netapp.spl.common.restclient.helper.PropertiesFileReader.AGENT_CONFIG_SNAPCENTER_SERVER_PORT;
import static com.netapp.spl.common.restclient.helper.PropertiesFileReader.AGENT_CONFIG_SNAPCENTER_SERVER_PROTOCOL;
import static com.netapp.spl.common.restclient.helper.PropertiesFileReader.SCCLI_CONFIG_SNAPCENTER_SERVER_CONNECTION_TIMEOUT_SECONDS;
import static com.netapp.spl.common.restclient.helper.PropertiesFileReader.SCCLI_CONFIG_SNAPCENTER_SERVER_RECEIVE_TIMEOUT_SECONDS;

import org.apache.commons.lang3.StringUtils;

import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;
import com.netapp.spl.common.restclient.helper.PropertiesFileReader;

/***
 * This class is for having SnapCenter Server related activity. It has an
 * utility for consuming snapcenter rest end point services, Polling server jobs
 * etc..
 *
 */
public class SnapCenterServerServiceFactory {

    private static final long DEFAULT_CONNECTION_TIMEOUT = 300 * 1000L;
    private static final long DEFAULT_RECEIVE_TIMEOUT = 600 * 1000L;

    private static ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(SnapCenterServerServiceFactory.class);

    private SnapCenterServerServiceFactory() {

    }

    /**
     * Initializing Service by reading spl.properties file.
     * 
     * @throws IllegalArgumentException
     * @throws IllegalStateException
     */
    public static SnapCenterServerServiceHelper getSnapCenterServerServiceHelperInstance() {

        PropertiesFileReader propertiesFileReader = PropertiesFileReader
                .getInstance();

        SnapCenterServerConfig snapCenterServerConfig = new SnapCenterServerConfig();

        // Initialize Snap Center Host
        String snapCenterHost = propertiesFileReader
                .getProperty(AGENT_CONFIG_SNAPCENTER_SERVER_HOST);
        if (StringUtils.isBlank(snapCenterHost)) {
            throw new IllegalStateException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_SNAPCENTER_SERVER_SERVICE_INIT_SERVER_HOST_EMPTY_ERROR));
        }

        // Initialize Snap Center Port
        int snapCenterPort;
        String snapCenterPortValue = propertiesFileReader
                .getProperty(AGENT_CONFIG_SNAPCENTER_SERVER_PORT);
        if (StringUtils.isBlank(snapCenterPortValue)) {
            throw new IllegalStateException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_SNAPCENTER_SERVER_SERVICE_INIT_SERVER_PORT_EMPTY_ERROR));
        }
        try {
            snapCenterPort = Integer.parseInt(snapCenterPortValue);
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException(
                    MessageBundle
                            .localizeMessage(
                                    CLIMsgKey.CLI_SNAPCENTER_SERVER_SERVICE_INIT_SERVER_PORT_INVALID_ERROR,
                                    snapCenterPortValue));
        }
        if (!isValidPortNumber(snapCenterPort)) {
            throw new IllegalArgumentException(
                    MessageBundle
                            .localizeMessage(
                                    CLIMsgKey.CLI_SNAPCENTER_SERVER_SERVICE_INIT_SERVER_PORT_INVALID_ERROR,
                                    snapCenterPortValue));
        }

        // Initialize Snap Center Protocol
        String snapCenterProtocol = propertiesFileReader
                .getProperty(AGENT_CONFIG_SNAPCENTER_SERVER_PROTOCOL);
        if (StringUtils.isBlank(snapCenterProtocol)) {
            throw new IllegalStateException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_SNAPCENTER_SERVER_SERVICE_INIT_SERVER_PROTOCOL_EMPTY_ERROR));
        }
        if (!isValidServerProtocol(snapCenterProtocol)) {
            throw new IllegalArgumentException(
                    MessageBundle
                            .localizeMessage(
                                    CLIMsgKey.CLI_SNAPCENTER_SERVER_SERVICE_INIT_SERVER_PROTOCOL_INVALID_ERROR,
                                    snapCenterProtocol));
        }

        long connectionTimeout = DEFAULT_CONNECTION_TIMEOUT;
        long recieveTimeout = DEFAULT_RECEIVE_TIMEOUT;

        // Get Connection timeout
        String connTimeOut = propertiesFileReader
                .getProperty(SCCLI_CONFIG_SNAPCENTER_SERVER_CONNECTION_TIMEOUT_SECONDS);
        if (StringUtils.isNotBlank(connTimeOut)) {
            try {
                long connectionTimeoutInSeconds = Long.parseLong(connTimeOut);
                connectionTimeout = connectionTimeoutInSeconds * 1000L;
                CLIUtils.writeToLog("Setting connection timeout to "
                        + connectionTimeoutInSeconds + " seconds", logger,
                        LogLevel.DEBUG);
            } catch (Exception e) {
                CLIUtils.writeToLog(e, logger, LogLevel.DEBUG);
            }
        } else {
            CLIUtils.writeToLog(
                    "Setting connection timeout to default value 300 seconds.",
                    logger, LogLevel.DEBUG);
        }

        // Get recieve timeout
        String recivTimeOut = propertiesFileReader
                .getProperty(SCCLI_CONFIG_SNAPCENTER_SERVER_RECEIVE_TIMEOUT_SECONDS);
        if (StringUtils.isNotBlank(recivTimeOut)) {
            try {
                long recieveTimeoutInSeconds = Long.parseLong(recivTimeOut);
                recieveTimeout = recieveTimeoutInSeconds * 1000L;
                CLIUtils.writeToLog("Setting recieve timeout to "
                        + recieveTimeoutInSeconds + " seconds", logger,
                        LogLevel.DEBUG);
            } catch (Exception e) {
                CLIUtils.writeToLog(e, logger, LogLevel.DEBUG);
            }
        } else {
            CLIUtils.writeToLog(
                    "Setting recieve timeout to default value 600 seconds.",
                    logger, LogLevel.DEBUG);
        }

        snapCenterServerConfig.setHost(snapCenterHost);
        snapCenterServerConfig.setProtocol(snapCenterProtocol);
        snapCenterServerConfig.setPort(snapCenterPort);
        snapCenterServerConfig.setConnectionTimeout(connectionTimeout);
        snapCenterServerConfig.setRecieveTimeout(recieveTimeout);
        SnapCenterServerServiceHelper snapCenterServerServiceHelper = new SnapCenterServerServiceHelper(
                snapCenterServerConfig);
        return snapCenterServerServiceHelper;

    }

    private static boolean isValidPortNumber(int portNumber) {
        return ((portNumber < 0 || portNumber > 65536) ? false : true);
    }

    private static boolean isValidServerProtocol(String protocol) {
        return protocol.equalsIgnoreCase("https")
                || protocol.equalsIgnoreCase("http");
    }

}
