/*
 * $Id: //depot/prod/orion/sc_main/mintaka/spl/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/startup/CLILauncher.java#1 $
 * 
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.snapcenter.cli.startup;

import com.netapp.common.logging.MessageBundle;

import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LogSettingsUtil;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.snapcenter.cli.command.AbstractSMSCommand;
import com.netapp.snapcenter.cli.command.CommandInterface;
import com.netapp.snapcenter.cli.command.HelpCommand;
import com.netapp.snapcenter.cli.command.parser.CommandParserFactory;
import com.netapp.snapcenter.cli.command.parser.CommandParserInterface;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;
import org.apache.commons.lang3.StringUtils;
import java.lang.reflect.Field;
import java.lang.reflect.Method;

public class CLILauncher {

    public static final String CLI_EXECUTABLE_NAME = MessageBundle
            .localizeMessage(CLIMsgKey.CLI_EXECUTABLE_NAME);

    private static ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(CLILauncher.class);

    private CLILauncher() {
    }


    public static void main(String[] args) {

    	disableAccessWarnings();
        LogSettingsUtil.setConfigurationSettings();

        int exitCode = 0;

        // Instantiate the parser
        CommandParserInterface commandParser = CommandParserFactory
                .getCommandParser();
        // Register All commands with parser
        commandParser.registerAllCommands();

        CommandInterface command;

        try {

            if (checkForCliHelp(args)) {
                CLIUtils.printCliUsage(commandParser);
                System.exit(exitCode);
            }

            CLIUtils.writeToLog(StringUtils.join(args, " "), logger,
                    LogLevel.DEBUG);

            CLIUtils.writeToLog(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_COMMAND_INVOCATION_STARTED),
                    logger, LogLevel.DEBUG);

            /**
             * Parse command line arguments
             */
            command = commandParser.parseCommand(args);

            /**
             * Run the command
             */
            if (command instanceof HelpCommand) {
                String commandName = ((HelpCommand) command)
                        .getCommandNameForHelp();
                if (commandName.isEmpty()) {
                    CLIUtils.printCliUsage(commandParser);
                } else {
                    commandParser.printUsage(commandName);
                }
            } else if (command instanceof AbstractSMSCommand) {
                if (((AbstractSMSCommand) command).isHelpSpecified()) {
                    commandParser.printUsage(command.getCommandName());
                } else {
                    command.run();
                }
            } else {
                command.run();
            }
        } catch (CLIException e) {
            CLIUtils.writeToLogAndConsole(e.getLocalizedMessage(), logger,
                    LogLevel.ERROR);
            CLIUtils.writeToLog(e, logger, LogLevel.DEBUG);

            /**
             * Check for null before printing. If Response code is null then the
             * exception might be due to some other error which will be printed
             * above in localized message. Else this is a Rest Call exception
             * and log respective http response code and status.
             */
            if (e.getResponseStatusCode() != null
                    && e.getResponseStatusMessage() != null) {
                CLIUtils.writeToLog(
                        "Server http response status: "
                                + e.getResponseStatusCode()
                                + ", Status message: "
                                + e.getResponseStatusMessage(), logger,
                        LogLevel.DEBUG);
            }
            exitCode = 1;
        } catch (NullPointerException e) {
            CLIUtils.writeToLogAndConsole(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_INTERNAL_UNEXPECTED_ERROR),
                    logger, LogLevel.ERROR);
            CLIUtils.writeToLog(e, logger, LogLevel.DEBUG);
            exitCode = 1;
        } catch (Exception e) {
            CLIUtils.writeToLogAndConsole(e.getLocalizedMessage(), logger,
                    LogLevel.ERROR);
            CLIUtils.writeToLog(e, logger, LogLevel.DEBUG);
            exitCode = 1;
        } finally {
            CLIUtils.closeResources();
            System.exit(exitCode);
        }
    }

    /**
     * The below code will disable --illegal-access warning errors that pop
     * up from JRE 9 onwards on illegal reflection done by Orika.
     * It is not good to throw this message to end user. Hence the below
     * function suppresses it.
     * Taken from:
     * https://stackoverflow.com/questions/46454995/how-to-hide-warning-illegal-reflective-access-in-java-9-without-jvm-argument
     */
    @SuppressWarnings("unchecked")
    public static void disableAccessWarnings() {
        try {
            Class unsafeClass = Class.forName("sun.misc.Unsafe");
            Field field = unsafeClass.getDeclaredField("theUnsafe");
            field.setAccessible(true);
            Object unsafe = field.get(null);

            Method putObjectVolatile = unsafeClass.getDeclaredMethod("putObjectVolatile", Object.class, long.class, Object.class);
            Method staticFieldOffset = unsafeClass.getDeclaredMethod("staticFieldOffset", Field.class);

            Class loggerClass = Class.forName("jdk.internal.module.IllegalAccessLogger");
            Field loggerField = loggerClass.getDeclaredField("logger");
            Long offset = (Long) staticFieldOffset.invoke(unsafe, loggerField);
            putObjectVolatile.invoke(unsafe, loggerClass, offset, null);
        } catch (Exception ignored) {
        }
    }
    
    /*
     * If nothing is specified or if only (-help || -?) is specified then
     * display snapcenter-cli usage.
     */
    private static boolean checkForCliHelp(String[] args) {
        if (args == null || args.length == 0) {
            return true;
        }
        if (args.length > 0) {
            String arg1 = args[0];
            if (arg1.equalsIgnoreCase("-help") || arg1.equalsIgnoreCase("-?")) {
                return true;
            }
        }
        return false;
    }

}
